
import unicodedata
import re

# Taken from Django
def slugify(value, allow_unicode=False):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    return re.sub(r'[-\s]+', '-', value)





import logging
logger = logging.getLogger('StegoVeritas:Modules:Multi')

import os
import sys
from .. import ModuleBase

class MultiHandler(ModuleBase):

    def __init__(self, veritas):
        super().__init__(veritas)

        # Given it's the mutli format handler, just assume we can handle it and let the plugins figure out the rest.
        self.valid = True
        
        """
        # Should we run all default plugins?
        if all([not veritas.args.imageTransform, not veritas.args.extractLSB, not veritas.args.bruteLSB, veritas.args.colorMap is None, veritas.args.colorMapRange is None, not veritas.args.trailing, not veritas.args.meta]):
            self._default_run = True
        else:
            self._default_run = False
        """

    @property
    def description(self):
        return ""





import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:Trailing')

import os
from struct import unpack
from .. import png
from ..gif import gif as gif_module

def run(image):
    """Extracts trailing data from the image.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR/trailing_data.bin
    """

    global output_file
    output_file = os.path.join(image.veritas.results_directory, "trailing_data.bin")

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.trailing:
        logger.debug('Nothing to do.')
        return

    try:
        if image.file.format == "JPEG":
            jpeg(image)
        elif image.file.format == "TIFF":
            tiff(image)
        elif image.file.format == "PNG":
            png(image)
        elif image.file.format == "BMP":
            bmp(image)
        elif image.file.format == "GIF":
            gif(image)
        else:
            print("Image Trailing: No support yet for format {0}".format(image.file.format))
            return
    except Exception as e:
        print("Image Trailing: Something went wrong... please submit a bug report. Error: {}".format(e))


def gif(image):
	
	# Load up the gif
	g = gif_module(fileName=image.veritas.file_name)
	
	# Parse it
	g.parse()
	
	# Check for excess info
	if len(g.gif) > 0:
		print("Discovered trailing data: {0}".format(g.gif))
		with open(output_file, "wb") as outFile:
			outFile.write(g.gif)


def png(image):
	pngFile = open(image.veritas.file_name,"rb").read()
	
	# TODO: This isn't 100% accurate. Rework to follow tags correctly.
	pngFile = pngFile.split(b"IEND")[1][4:]
	
	if len(pngFile) != 0:
		print("Discovered Trailing Data:\n{0}".format(pngFile))
		with open(output_file,"wb") as outFile:
			outFile.write(pngFile)


def bmp(image):
	# This one is pretty easy
	myBMP = open(image.veritas.file_name,"rb").read()
	
	size = unpack("<I",myBMP[2:6])[0]
	
	if len(myBMP[size:]) != 0:
		print("Discovered Trailing Data:\n{0}".format(myBMP[size:]))
		with open(output_file,"wb") as outFile:
			outFile.write(myBMP[size:])


def tiff(image):
	
	# Read in the file
	with open(image.veritas.file_name,"rb") as myFile:
		steg = myFile.read()
	
	# Parse magic
	if steg[0:4] == b"II*\x00":
		fmt = "<I"
		fmt_s = "<h"
	elif steg[0:4] == b"MM*\x00":
		fmt = ">I"
		fmt_s = ">h"
	else:
		print("Trailing: Error Invalid tiff magic numbers")
		return
	
	# Read header
	ifd = unpack(fmt,steg[4:8])[0]
	
	# Read the number of tags
	nEntries = unpack(fmt_s,steg[ifd:ifd+2])[0]
	
	# We want to find the maximum address
	# Right now our max address is the end of the IFD block
	maxAddr = ifd + (nEntries*0xc) + 2
	
	# Loop through the tags
	for i in range(nEntries):
		# Figure out our current file location
		curAddr = ifd + 2 + (i * 0xc)

		tag = unpack(fmt_s,steg[curAddr:curAddr+2])[0]
		tagType = unpack(fmt_s,steg[curAddr+2:curAddr+4])[0]
		count = unpack(fmt,steg[curAddr+4:curAddr+8])[0]
		
		# print("Tag: {0}\nType: {1}".format(tag,tagType))

		# Tag types of ASCII and Unknown both have offsets associated
		if tagType == 2 or tagType == 7:
			offset = unpack(fmt,steg[curAddr+8:curAddr+0xc])[0]
			# See if we have a new winner
			if (offset + count) > maxAddr:
				maxAddr = offset + count
				#print("New max offset")
			# print("Found new offset: {0}".format(hex(offset)))

	# See if we have data hiding at the end
	if len(steg) > maxAddr:
		print("Trailing Data Discovered... Saving")
		print(steg[maxAddr:])
		with open(output_file,"wb") as outFile:
			outFile.write(steg[maxAddr:])


def jpeg(image):

	# Official specs here: http://www.w3.org/Graphics/JPEG/itu-t81.pdf	

	# Index for marching through the file	
	i = 0
	
	# These markers don't have a length attribute
	nonLenMarkers = [ b'\xff\xd8', b'\xff\x01', b'\xffd0', b'\xffd1', b'\xffd2', b'\xffd3', b'\xffd4', b'\xffd5', b'\xffd6', b'\xffd7' ]

	# Open up the file
	with open(image.veritas.file_name,"rb") as myFile:
		steg = myFile.read()
	
	while True:
		# Grab the current header
		hdr = steg[i:i+2]
		
		# TODO: Add py logging here
		#print("Found Header: {0}".format(hdr))
		
		# if Start of Image, Temporary Private, Restart, things that don't have an associated length field
		if hdr in nonLenMarkers:
			# Just move to the next marker
			i = i + 2
			continue
		
		# If we've found our way to the end of the jpeg
		if hdr == b'\xff\xd9':
			#print("Made it to the end!")
			# Increment 2 so we can check the length
			i += 2
			break
		
		# Unpack the length field
		ln = unpack(">H",steg[i+2:i+4])[0]
		
		# print("Found Length: {0}".format(ln))
		
		# Update the index with the known length
		i = i+ln+2
		
		# When we hit scan data, we scan to the end of the format
		if hdr == b'\xff\xda':
			#print("Start of Scan data")
			# Find the end marker
			i += steg[i:].index(b'\xff\xd9')
	
	# Check for trailers
	if i != len(steg):
		print("Trailing Data Discovered... Saving")
		print(steg[i:])
		# Save it off for reference
		with open(output_file, "wb") as outFile:
			outFile.write(steg[i:])




import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:BruteLSB')

import os
import shutil

def run(image):
    """Attempts to brute force out any LSB stego.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR
    """

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.bruteLSB:
        logger.debug('Nothing to do.')
        return

    if image.file.mode == "P":
        logger.debug('Image is colormap, skipping..')
        return

    # Try all the same indexes first. More likely stuff up front.
    # i.e.: 0,0,0  1,1,1,  2,2,2
    for i in range(0,8):
        print("Trying {0}.{0}.{0}".format(i))
        o = image.dumpLSBRGBA(red_index=[i],green_index=[i],blue_index=[i])
        image.veritas.test_output(o)

    # Try Red
    i = []
    for x in range(8):
        i.append(x)
        print("Trying Red {0}".format(i))
        o = image.dumpLSBRGBA(red_index=i)
        image.veritas.test_output(o)

    # Try Green
    i = []
    for x in range(8):
        i.append(x)
        print("Trying Green {0}".format(i))
        o = image.dumpLSBRGBA(green_index=i)
        image.veritas.test_output(o)

    # Try Blue
    i = []
    for x in range(8):
        i.append(x)
        print("Trying Blue {0}".format(i))
        o = image.dumpLSBRGBA(blue_index=i)
        image.veritas.test_output(o)

    # Try Alpha
    if "A" in image.file.mode:
        i = []
        for x in range(8):
            i.append(x)
            print("Trying Alpha {0}".format(i))
            o = image.dumpLSBRGBA(alpha_index=i)
            image.veritas.test_output(o)

    # Try across the board
    # i.e.: 0,0,0  01,01,01 etc
    i = []
    for x in range(8):
        i.append(x)
        print("Trying {0}x{0}x{0}".format(i))
        o = image.dumpLSBRGBA(red_index=i,green_index=i,blue_index=i)
        image.veritas.test_output(o)

    return
    for r in range(0,8):
        for g in range(0,8):
            for b in range(0,8):
                print("Trying {0}.{1}.{2}".format(r,g,b))
                o = image.dumpLSBRGBA(red_index=[r],green_index=[g],blue_index=[b])
                image.veritas.test_output(o)



import os.path,os

# Get our script dir
SCRIPTDIR = os.path.dirname(os.path.realpath(__file__))

# Get the current dir
PWD = os.getcwd()

TEMPFILE = os.path.join(SCRIPTDIR,"stegoveritas.temp")
# TODO: Use python to make this dynamic
RESULTSDIR = os.path.join(PWD,"results")





import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:Filters')

import os
from PIL import Image,ImageFilter
import numpy
from copy import copy

def run(image):
    """Performs various graphical filters on the image.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR
    """

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.imageTransform:
        logger.debug('Nothing to do.')
        return

    # Set alpha if it exists.. f_a == the same image but with the alpha layer full
    bands = image.file.getbands()
    if "A" in bands:
            index = bands.index("A")
            f_a = numpy.array(image.file)
            f_a[:,:,index] = 0xff
            f_a = Image.fromarray(f_a)
    # Using an else here to keep things simple. If you use f_a you will always have a working non-alpha image
    else:
            f_a = image.file

    # Check what type of image is it
    # ColorMap
    if "P" in bands:
        logger.debug('Image is colormap, skipping.')
        return

    ###########
    # Filters #
    ###########
    
    # Built-in filters to apply
    filters = [ImageFilter.EDGE_ENHANCE, ImageFilter.EDGE_ENHANCE_MORE, ImageFilter.FIND_EDGES, ImageFilter.MaxFilter, ImageFilter.MedianFilter, ImageFilter.MinFilter, ImageFilter.ModeFilter, ImageFilter.SHARPEN, ImageFilter.SMOOTH, ImageFilter.GaussianBlur]
    
    for filt in filters:
            g = f_a.filter(filt)
            g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_" + filt.name.replace(" ","_") + ".png"))
    
    ################
    # Color planes #
    ################
    g = numpy.array(f_a)
    g[:,:,2] *= 0
    g[:,:,1] *= 0
    g = Image.fromarray(g)
    g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_red_plane.png"))
    
    g = numpy.array(f_a)
    g[:,:,2] *= 0
    g[:,:,0] *= 0
    g = Image.fromarray(g)
    g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_green_plane.png"))

    g = numpy.array(f_a)
    g[:,:,0] *= 0
    g[:,:,1] *= 0
    g = Image.fromarray(g)
    g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_blue_plane.png"))

    # TODO: Still assuming all the positions for the given colors in the tuple
    if "A" in bands:
            g = numpy.array(image.file)
            g[:,:,0] *= 0
            g[:,:,1] *= 0
            g[:,:,2] *= 0
            g = Image.fromarray(g)
            g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_alpha_plane.png"))

    
    ##################
    # Bit Map Planes #
    ##################
    colorPlanes = image.file.split()
    
    # TODO: This is an assumption on ordering... Maybe find a better way.
    color = "Red"
    for plane in colorPlanes:
            for i in range(8):
                    temp = numpy.array(plane)
                    # Shift over to get right offset
                    temp = temp >> i
                    # Get the bit
                    temp = temp & 1
                    # Fix-up the background
                    temp[temp == 1] = 255
                    # Save it off
                    temp = Image.fromarray(temp)
                    temp.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_{0}_{1}.png".format(color,i)))
            # Move to next color
            if color == "Red":
                    color = "Green"
            elif color == "Green":
                    color = "Blue"
            elif color == "Blue":
                    color = "Alpha"
            elif color == "Alpha":
                    color = "Error"
            else:
                    raise Exception("We got more planes than we were expecting...\nPlanes: {0}".format(colorPlanes))

    ##########
    # Invert #
    ##########
    g = numpy.array(f_a)
    g = g ^ 0xff
    g = Image.fromarray(g)
    g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_inverted.png"))






#!/usr/bin/env python3

from . import Colorer

import logging
logging.basicConfig(level=logging.WARN)

logger = logging.getLogger('StegoVeritas')

import binascii
import os, os.path
import argparse
from .config import *
from .version import version

import magic
import time
import shutil
from copy import copy

from .install_deps import required_packages

class StegoVeritas(object):

    def __init__(self, args=None):
        """
        Args:
            args (list, optional): Arguments as if you passed them via the command line (i.e.: ['-out','directory','-meta'])
        """
        
        self._preflight()
        self._parse_args(args)
        self.modules = []


    def run(self):
        """Run analysis on the file."""
        for module in modules.iter_modules(self):
            print('Running Module: ' + module.__class__.__name__)
            print(module.description)
            module.run()
            self.modules.append(module)

    def test_output(self, thing):
            """
            Args:
                thing (bytes): Renerally from the dump functions
                    ex: thing = b'\x01\x02\x03'

            Returns:
                Nothing. Move output into keep directory if it's worth-while    

            
            Test if output is worth keeping. If it is, move it into the results directory.
            Initially, this is using the Unix file command on the output and checking for non "Data" returns
            """
            
            assert type(thing) == bytes, 'test_output got unexpected thing type of {}'.format(type(thing))

            # TODO: Test new logic...
            # TODO: Iterate through binary offset to find buried data

            m = magic.from_buffer(thing,mime=True)

            # Generic Output
            if m != 'application/octet-stream':
                m = magic.from_buffer(thing,mime=False)
                print("Found something worth keeping!\n{0}".format(m))
                # Save it to disk
                # TODO: Minor race condition here if we end up multi-processing
                with open(os.path.join(self.results_directory, str(time.time())), "wb") as f:
                    f.write(thing)
            
            # TODO: Check if strings of output contain a known word, save if so.

    def _preflight(self):
        """Checks for missing requirements."""
        
        missing_packages = []

        for package in required_packages:
            if shutil.which(package) is None:
                missing_packages.append(package)

        try:
            import libxmp.utils
        except:
            missing_packages.append('libexempi3')

        if missing_packages != []:
            logger.error('Missing the following required packages: ' + ', '.join(missing_packages))
            logger.error('Either install them manually or run \'stegoveritas_install_deps\'.')

    def _parse_args(self, args=None):

        parser = argparse.ArgumentParser(description='Yet another Stego tool',
                epilog = 'Have a good example? Wish it did something more? Submit a ticket: https://github.com/bannsec/stegoVeritas')

        # Core Options
        parser.add_argument('-out',metavar='dir',type=str, help='Directory to place output in. Defaults to ./results',default=os.path.abspath('./results'))
        parser.add_argument('-debug', action='store_true', help='Enable debugging logging.')
        parser.add_argument('file_name',metavar='file',type=str, default=False, help='The file to analyze')

        # Image Options
        image = parser.add_argument_group('image options')
        image.add_argument('-meta',action='store_true',help='Check file for metadata information')
        image.add_argument('-imageTransform',action='store_true',help='Perform various image transformations on the input image and save them to the output directory')
        image.add_argument('-bruteLSB',action='store_true',help='Attempt to brute force any LSB related stegonography.')
        image.add_argument('-colorMap',nargs="*",metavar='N',type=int, default=None, help='Analyze a color map. Optional arguments are colormap indexes to save while searching')
        image.add_argument('-colorMapRange',nargs=2,metavar=('Start','End'),type=int, default=None, help='Analyze a color map. Same as colorMap but implies a range of colorMap values to keep')
        image.add_argument('-extractLSB',action='store_true',help='Extract a specific LSB RGB from the image. Use with -red, -green, -blue, and -alpha')
        image.add_argument('-red',nargs='+',metavar='index',type=int)
        image.add_argument('-green',nargs='+',metavar='index',type=int)
        image.add_argument('-blue',nargs='+',metavar='index',type=int)
        image.add_argument('-alpha',nargs='+',metavar='index',type=int)
        image.add_argument('-trailing',action='store_true',help='Check for trailing data on the given file')

        # Multi Options
        multi = parser.add_argument_group('multi options')
        multi.add_argument('-exif',action='store_true',default=False,help='Check this file for exif information.')
        multi.add_argument('-xmp',action='store_true',default=False,help='Check this file for XMP information.')

        self.args = parser.parse_args(args)

        if self.args.debug:
            logging.root.setLevel(logging.DEBUG)

        self.file_name = self.args.file_name
        self.results_directory = self.args.out

        # Should this be considered an 'auto' run?
        # TODO: This is SUPER hacky... Should probably find a better way to determine if this is an auto run or not.
        auto = copy(self.args.__dict__)
        auto.pop('out')
        auto.pop('file_name')
        auto.pop('debug')
        self.args.auto = all(option in [False, None, []] for option in auto.values())
    
    ##############
    # Properties #
    ##############

    @property
    def file_name(self) -> str:
        return self.__file_name

    @file_name.setter
    def file_name(self, file_name: str) -> None:
        
        full_path = os.path.abspath(file_name)
        
        if not os.path.exists(full_path):
            logger.error('Cannot find file "{}"'.format(full_path))
            exit(1)

        logger.info('Analyzing file: ' + full_path)
        self.__file_name = full_path

    @property
    def results_directory(self) -> str:
        return self.__results_directory

    @results_directory.setter
    def results_directory(self, results_directory: str) -> None:
        full_path = os.path.abspath(results_directory)

        if os.path.exists(full_path) and not os.path.isdir(full_path):
            logger.error('Output path exists and is not a directory.')
            exit(1)

        os.makedirs(full_path, exist_ok=True)
        
        logger.info('Results Directory: ' + full_path)
        self.__results_directory = results_directory

    @property
    def modules(self) -> list:
        """list: List of all modules that have been run. NOTE: This will only be populated AFTER 'run' has been called and the modules themselves have been run."""
        return self.__modules

    @modules.setter
    def modules(self, modules):
        self.__modules = modules


def main(args=None):
    veritas = StegoVeritas(args=args)
    veritas.run()

from . import modules

if __name__ == '__main__':
    main()




from . import Colorer
import logging
logger = logging.getLogger('StegoVeritas:InstallDeps')

import platform
import subprocess

import getpass

# What is required by stegoveritas?
required_packages = ['binwalk', 'exiftool']

def main():

    dist_name, dist_version, dist_id = platform.dist()
    
    if dist_name.lower() == 'ubuntu':
        ubuntu()

    else:
        logger.error('Unhandled distribution to install deps: {}'.format(', '.join(platform.dist())))
        logger.error('Please poke me or submit a PR.')
        return

def ubuntu():
    
    packages = ['binwalk', 'libimage-exiftool-perl', 'libexempi3']

    subprocess.run(command_start + ['apt-get','update'])
    subprocess.run(command_start + ['apt-get','install','-y'] + packages)


# Standardize using sudo or not
command_start = ['sudo'] if getpass.getuser() != 'root' else []

if __name__ == '__main__':
    main()



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os, sys, ast

here = os.path.abspath(os.path.dirname(__file__))
long_description = "See website for more info."

# Importing during setup causes dependency issues
with open('stegoveritas/version.py') as f:
    exec(f.read())

# TODO: Get 'six' package up to newest version. Right now py010parser is forcing it to 1.10.0, which isn't great.
# https://github.com/d0c-s4vage/py010parser/pull/19

setup(
    name='stegoveritas',
    version=version,
    description='General Steganography detection tool.',
    long_description=long_description,
    url='https://github.com/bannsec/stegoVeritas',
    author='Michael Bann',
    author_email='self@bannsecurity.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'Environment :: Console'
    ],
    keywords='steg stego steganography stegoveritas',
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'dist']),
    install_requires=['pillow', 'numpy', 'pfp', 'python-magic', 'prettytable', 'exifread', 'python-xmp-toolkit'],
    extras_require={
        'dev': ['six==1.10.0','ipython','twine','pytest','python-coveralls','coverage','pytest-cov','pytest-xdist','sphinxcontrib-napoleon', 'sphinx_rtd_theme','sphinx-autodoc-typehints', 'pyOpenSSL'],
    },
    entry_points={
        'console_scripts': [
            'stegoveritas = stegoveritas.stegoveritas:main',
            'stegoveritas_install_deps = stegoveritas.install_deps:main',
        ],
    },
)





import logging
logger = logging.getLogger('StegoVeritas:Modules:Image')

import os
import sys
from PIL import Image
from prettytable import PrettyTable

from .. import ModuleBase
import hashlib

class SVImage(ModuleBase):

    def __init__(self, veritas):
        super().__init__(veritas)

        # Can we handle this file?
        try:
            self.file = Image.open(veritas.file_name)
            self.valid = True

        except:
            self.valid = False
            logger.info('Cannot handle this file type.')
            return
        
    # Change this to a primitive to dump any given index of a given color
    # Then, handle the weaving of those together in a different function
    def dumpLSBRGBA(self, red_index = [],green_index = [],blue_index = [],alpha_index = []):
        """
        Input: 
                red_index, green_index, blue_index, alpha_index as array of integer indexes (up to 8) to dump
                ex: [0],None,None would dump only the least significant bit of the Red field
        Action:
                Creates a byte array containing the output of the LSB dump (RGBA order) requested
                If needed, it will use the least significant bit first, then bit plane order of red->green->blue->alpha
        Returns:
                Byte array of the result of the action
                ex: b'\x01\x02\x03\x04' etc
        """


        ##################
        # Combine Output #
        ##################
        # We'll be keeping the binary string here
        binStr = ''

        # Figure out valid index ranges
        indexes = list(set(red_index + green_index + blue_index + alpha_index))
        indexes.sort()

        # Figure out what we have to work with
        bands = self.file.getbands()

        # Get the image bytes
        fBytes = self.file.tobytes()

        # TODO: The following assumes an ordering of RGBA. If this is ever not the case, things will get mixed up
        # Loop through all the bytes of the image
        for byte in range(0,self.file.size[0] * self.file.size[1] * len(bands),len(bands)):
                # Loop through all the possible desired indexes
                for index in indexes:
                        # If this is a value we're extracting
                        if index in red_index:
                                binStr += str(fBytes[byte + 0] >> index & 1)
                        if index in green_index:
                                binStr += str(fBytes[byte + 1] >> index & 1)
                        if index in blue_index:
                                binStr += str(fBytes[byte + 2] >> index & 1)
                        if index in alpha_index:
                                binStr += str(fBytes[byte + 3] >> index & 1)

        # Parse those into bytes
        bArray = []
        for i in range(0,len(binStr),8):
                bArray.append(int(binStr[i:i+8],2))

        # Change bytes into a bit array for writing
        bits = ''.join([chr(b) for b in bArray]).encode('iso-8859-1')

        return bits

    @staticmethod
    def hash_file(file_name):
        """Custom image file hasing algorithm in an attempt to make hashing less brittle to meta information.

        Args:
            file_name (str): String path to file
    
        Returns:
            bytes: Hash of the given file.
        """
        
        img = Image.open(file_name)
        f = img.tobytes()
        
        # Need to account for the palette
        if img.mode == 'P':
            f += img.palette.mode.encode()
            f += img.palette.palette

        img.close()
        return hashlib.sha256(f).hexdigest()
            
    
    @property
    def file(self):
        """PIL Image Instantiation."""
        return self.__file

    @file.setter
    def file(self, file):
        self.__file = file

    @property
    def description(self):
        table = PrettyTable(['Image Format', 'Mode'])
        table.add_row([self.file.format_description, 'ColorMap' if self.file.mode == 'P' else self.file.mode])
        return str(table)


def autoAnalysis(f,args):
	"""
	Input:
		f -- PIL file Image object
		args -- argument array as parsed by argparse
	Action:
		Perform a default set of analysis on the image
	Returns:
		Nothing
	"""
	
	# Check metadata
	print("Checking Meta Data\n")
	modules.image.imageMeta.auto(f,args)	
	
	# Check for trailing data
	print("Checking for trailing data")
	modules.image.imageTrailing.auto(f,args)
	
	# Run the filers on it
	print("Running image filters")
	modules.image.imageFilters.auto(f,args.outDir)

	if f.mode == "P":
		print("Image is a colormap, skipping LSB Extract")
		return

	# Brute LSB
	print("Attempting to brute force LSB items")
	modules.image.imageLSB.auto(f,args)
	
	
def run(fArray,args):
	"""
	Primary handler for Image type stego
	Input:
		fArray is the array returned by the previous open call. In this case it should be a PIL image object
		args is the array parsed by the main program and contains what arguments were passed to the program
	Action:
		Runs analysis on the Image, prompting the user if necessary
	Returns:
		Nothing
	"""
	
	# Grab the file object
	f = fArray[0]
	
	# Determine if we're dealing with the Alpha band
	hasAlpha = "A" in f.getbands()

	# Check for the default case
	if len(sys.argv) == 2:
		autoAnalysis(f,args)
		return

	# Check for action based on flags	
	if args.extractLSB:
		extractLSB(f,args)

	if args.bruteLSB:
		modules.image.imageLSB.auto(f,args)

	if args.colorMapRange != None:
		args.colorMap = range(args.colorMapRange[0],args.colorMapRange[1]+1)

	if args.colorMap != None:
		modules.image.imageFilters.colorMap(f,args.outDir,args.colorMap)
	
	if args.imageTransform:
		modules.image.imageFilters.auto(f,args.outDir)
	
	if args.meta:
		modules.image.imageMeta.auto(f,args)

	if args.trailing:
		modules.image.imageTrailing.auto(f,args)

#from . import imageLSB, imageFilters, imageMeta, imageTrailing




import logging
logger = logging.getLogger('StegoVeritas:Modules')

import importlib
import inspect
import os

class ModuleBase(object):
    """Define things that need to be in the module class."""

    def __init__(self, veritas):
        """
        Args:
            veritas (stegoveritas.StegoVeritas): Base StegoVeritas instantiated class.
        """
        self.veritas = veritas
        self.valid = False

    def run(self):
        """By default, this will dynamically load any python files under the 'analysis' subfolder and call the 'run' method with itself as the only argument."""

        # List out the analysis modules
        module_base = str(self.__class__.__module__)
        analysis_dir = os.path.join(os.path.dirname(inspect.getfile(self.__class__)), 'analysis')
        _,_,files = next(os.walk(analysis_dir))

        for f in files:
            if f.endswith('.py') and f != '__init__.py':
                module = importlib.import_module('.analysis.' + f[:-3], module_base)
                module.run(self)

    @property
    def veritas(self):
        return self.__veritas

    @veritas.setter
    def veritas(self, veritas):
        self.__veritas = veritas

    @property
    def valid(self):
        """Bool: Is this class valid for the given file provided?"""
        return self.__valid

    @valid.setter
    def valid(self, valid):
        assert type(valid) is bool, 'Unexpected type for valid of {}'.format(type(valid))
        self.__valid = valid

    @property
    def description(self) -> str:
        """str: Description of things we know about this file."""
        return 'This hasn\'t been implemented yet.'

def iter_modules(veritas):
    """Iterate through handlers until one of them loads correctly or we run out of handlers.

    Returns:
        Valid handler instantiation or None.
    """

    for module in modules:
        instance = module(veritas)
        
        if instance.valid:
            yield instance

from .image import SVImage
from .multi import MultiHandler

# List of classes of modules to try
modules =  [ SVImage, MultiHandler ]
